#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "booleen.h"
#include "stk.h"

#include "traducteur.h"

/* Prototypes */

static char *Suppresion_Blancs(char *Chaine);
static char *ltoa(long l_to_a);
static char *dtoa(double d_to_a);
static char *Lit_Nom_Fichier(char **nf);

static char *Reconnait_Cellule(char **cell);
static char *Reconnait_Fonction(char **fct);
static char *Reconnait_Nombre(char **nbr);
static char *Reconnait_Symbole(char **symb);
static char *Reconnait_Chaine(char **chaine);

static char *Analyse_Prochain_Type(char **c);
static char *STk_traducteur(char *Chaine_C);

static BOOLEEN Cellule_Figee(char car);
static BOOLEEN Cellule_Externe(char car);
static BOOLEEN Lettre(char car);
static BOOLEEN Chiffre(char car);
static BOOLEEN Symbole_Special(char car);
static BOOLEEN Chaine(char car);

/* Fonctions / Procdures */

static char *Suppresion_Blancs(char *Chaine) {

  char *Temp = calloc(strlen(Chaine), sizeof(char));
  char *Chaine_ret = Temp;

  while(*Chaine) {

    /* Si celui qui lit ce commentaire connait un truc plus simple
       qu'il ne se gene pas et qu'il change ! */

    /* Si c'est pas un blanc, on recopie */
    if ((*Chaine != 8) && (*Chaine != 32)) {

      /* Et en plus on fait de la mise en majuscule ! */
      if((*Chaine >= 'a') && (*Chaine <= 'z')) {
        *Temp = (*Chaine++);
        *Temp++ -= ('a' - 'A');
      }
      else {
        (*Temp++) = (*Chaine++);
      }
    }
    /* sinon, on saute les blancs */
    else {
      Chaine ++;
    }
  }
  return Chaine_ret;
}

/* Quelques prdicats de reconnaissance de la chaine tape */

static BOOLEEN Cellule_Figee(char car) {
  if (car == FIGEE)
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Cellule_Externe(char car) {
  if (car == DEBUT_NOM_FICHIER)
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Lettre(char car) {
  if ((car >= 'A') && (car <= 'Z'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Chiffre(char car) {
  if ((car >='0') && (car <= '9'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Symbole_Special(char car) {
  if ((car == SEPARATEUR) ||
      (car == DEF_ZONE) ||
      (car == '(') ||
      (car == ')') ||
      (car == '+') ||
      (car == '-') ||
      (car == '*') ||
      (car == '/') ||
      (car == '%'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Chaine(char car) {
  if (car == '"')
    return VRAI;
  else
    return FAUX;
}

/* Les fonctions de "traduction" des diffrents types connus */

static char *ltoa(long l_to_a) {

  char *ret_a = calloc(100, sizeof(char));

  sprintf(ret_a, "%ld ", l_to_a);

  return ret_a;
}

static char *dtoa(double d_to_a) {

  char *ret_a = calloc(100, sizeof(char));

  sprintf(ret_a, "%g ", d_to_a);

  return ret_a;
}

static char *Lit_Nom_Fichier(char **nf) {
  char *pt = *nf;

  char *ret_nf = calloc(255, sizeof(char));
  char *ret = ret_nf;

  pt++;

  *ret_nf++ = '"';

  while((*pt) && (! Cellule_Externe(*pt)))
         *ret_nf++ = *pt++;

  if(! (*pt))
    return "'erreur";

  *ret_nf++ = '"';
  *ret_nf   = ' ';

  *nf = ++pt;

  return ret;
}

static char *Reconnait_Cellule(char **cell) {

  long Colonne = 0;
  long Ligne = 0;

  char *Ret_Cell = calloc(200, sizeof(char));

  char *pt_cell = *cell;

  BOOLEEN Ferme_Feuille = FAUX;

  /* On reconnait d'abord le nom de la feuille externe s'il y a lieu */

  if(Cellule_Externe(*pt_cell)) {
    strcat(Ret_Cell, "(list 'Feuille ");
    strcat(Ret_Cell, Lit_Nom_Fichier(cell));
    pt_cell = *cell;
    Ferme_Feuille = VRAI;
  }

  /* On regarde si la colonne de la cellule est fige */

  strcat(Ret_Cell, "(list ");

  if(Cellule_Figee(*pt_cell)) {
    strcat(Ret_Cell, "'$ ");
    pt_cell++;
    *cell = pt_cell;
  }
  else
    strcat(Ret_Cell, "'r ");

  /* Reconnaissance de la cellule */

  /* Colonne */

  while((*pt_cell) && Lettre(*pt_cell)) {
    Colonne *= 26;
    Colonne += ((*pt_cell++) - 'A' + 1);
  }

  *cell = pt_cell;

  if(Colonne>0)
         strcat(Ret_Cell, ltoa(Colonne));
  else
         strcat(Ret_Cell, "'erreur");

  /* On regarde si la ligne de la cellule est fige */

  if(Cellule_Figee(*pt_cell)) {
    strcat(Ret_Cell, "'$ ");
    pt_cell++;
    *cell = pt_cell;
  }
  else
    strcat(Ret_Cell, "'r ");

  /* Ligne */
  
  while((*pt_cell) && Chiffre(*pt_cell)) {
    Ligne *= 10;
    Ligne += (*pt_cell++) - '0';
  }

  *cell = pt_cell;

  if(Ligne>0)
         strcat(Ret_Cell, ltoa(Ligne));
  else
         strcat(Ret_Cell, "'erreur");

  strcat(Ret_Cell, ")");

  if (Ferme_Feuille)
         strcat(Ret_Cell, ")");

  return Ret_Cell;
}

static char *Reconnait_Fonction(char **fct) {

  char *pt_nfct = *fct;

  char *Ret_Fct = calloc(1000, sizeof(char));

  char *Ret_Tmp = Ret_Fct;

  strcat(Ret_Fct, "(list '");
  
  Ret_Fct += strlen(Ret_Fct);

  /* Lecture du symbole de la fonction */

  while(Lettre(*pt_nfct))
         *Ret_Fct++ = *pt_nfct++;

  *Ret_Fct++ = ' ';

  Ret_Fct = Ret_Tmp;
  
  *fct = pt_nfct + 1;

  if(*pt_nfct!='(')
         strcat(Ret_Fct, "'erreur)");
  else {
         do {
	   strcat(Ret_Fct, Analyse_Prochain_Type(fct));
         } while((**fct) && (**fct) != ')');

         strcat(Ret_Fct, " )");
         (*fct)++;
  }
  return Ret_Fct;
}

static char *Reconnait_Nombre(char **nbr) {
  char *pt_nb = *nbr;

  double Nombre = 0.;
  double Puissance = 1.;

  while(Chiffre(*pt_nb)) {
         Nombre *= 10;
         Nombre += (*pt_nb++) - '0';
  }

  if((*pt_nb)==DECIMALE) {
    pt_nb++;
    while(Chiffre(*pt_nb)) {
      Puissance *= 10;
      Nombre *= 10;
      Nombre += (*pt_nb++) - '0';
    }
    Nombre /= Puissance;
  }
  *nbr = pt_nb;

  return dtoa(Nombre);
}

static char *Reconnait_Symbole(char **symb) {
  
  char car = **symb;

  char *Ret_Symb = calloc(5, sizeof(char));

  Ret_Symb[0] = ' ';
  Ret_Symb[1] = '\'';
  Ret_Symb[2] = car;
  Ret_Symb[3] = ' ';

  (*symb)++;

  return Ret_Symb;
}

static char *Reconnait_Chaine(char **chaine) {

  char *Ret_Chaine = calloc(255, sizeof(char));
  char *pt_chaine = *chaine;
  char *pt_ret = Ret_Chaine;

  *Ret_Chaine++ = *pt_chaine++;

  while(!Chaine(*pt_chaine))
    *Ret_Chaine++ = *pt_chaine++;

  *Ret_Chaine = *pt_chaine++;

  *chaine = pt_chaine;

  return pt_ret;
}

/* Fonction d'analyse du prochaine type */

static char *Analyse_Prochain_Type(char **c) {

  char *pt_c = *c;
  char *pt_t;

  /* Cellule ? */

  if((Cellule_Figee(*pt_c)) ||  /* A priori un $ */
     (Cellule_Externe(*pt_c)))  /* une ' pour le nom de la feuille */
    return Reconnait_Cellule(c);

  /* Cellule ou fonction ? */

  if(Lettre(*pt_c)) { /* Soit fonction, soit cellule */
    pt_t = pt_c;
    while(Lettre(*pt_t)) pt_t++;
    if(Chiffre(*pt_t)||Cellule_Figee(*pt_t))
      return Reconnait_Cellule(c);
    else
      return Reconnait_Fonction(c);
  }

  /* Nombre ? */

  if((Chiffre(*pt_c)) || (*pt_c) == DECIMALE)
    return Reconnait_Nombre(c);

  /* Symbole ? */

  if(Symbole_Special(*pt_c))
    return Reconnait_Symbole(c);

  /* Chaine de caractres */

  if(Chaine(*pt_c))
    return Reconnait_Chaine(c);
    
  /* Type non reconnu */

  return "'erreur";

}

static char *STk_traducteur(char *Chaine_C) {
  char *PT;

  /* Chaine de travail */
  char *Chaine_T;

  /* Chaine servant  l'valuation du rsultat */
  char *Chaine_R = calloc(1000, sizeof(char));

  /* Analyse de la chaine */

  /* Suppression des blancs et recopie de Chaine_C : Raciste ! */

  Chaine_T = Suppresion_Blancs(Chaine_C);

  /* On remplit le dbut de la chaine qui sert au rsultat */

  strcpy(Chaine_R, "(list ");

  /* On regarde d'abord s'il y a quelque chose a faire ! */

  if (*Chaine_T != '=') {

    /* On met 'no_eval dans la liste retourne
       pour que le moteur sache qu'il n'y a rien a faire */

    strcat(Chaine_R, "'no_eval ");
         strcat(Chaine_R, "\"");
    strcat(Chaine_R, Chaine_T);
         strcat(Chaine_R, "\")");
    return Chaine_R;
  }

  /* C'est bien un gal, on doit analyser la suite */

  Chaine_T ++;
  do {
    PT = Analyse_Prochain_Type(&Chaine_T);

    strcat(Chaine_R, PT);
  } while(((*Chaine_T) != '\0') && (strcmp("'erreur", PT) != 0));

  strcat(Chaine_R, ")");

  return Chaine_R;
}

void main(int argc,char *argv[]) {

  printf ("\nEvaluation de :");
  printf ("%s\n", argv[1]);
  printf ("\nResultat :\n");
  printf ("%s\n", STk_traducteur (argv[1]));
}

PRIMITIVE STk_init_traducteur(void) {

  STk_add_new_primitive("traducteur", tc_subr_1, STk_traducteur);

  return UNDEFINED;
}

/* La grand-mere est la suivante :

   Cellule :

     \'NomFic\'Ref
   | Ref

   Ref :

     ['$']Lettre+['$']Chiffre+

   Fonction :

     Lettre+'('Appel')'

   Nombre :

     Chiffre+

   Chaine :

     \".\"

   Divers :

     ; | : | * | / | - | + | %

*/
