/*
 * t r a d u c t e u r . c
 *
 *  xxl project -- Universite de Nice Sophia Antipolis
 *                 U.F.R. Sciences - Maitrise d'Informatique
 *                 project supervisor  V. Granet (vg@unice.fr)
 * 
 *  Copyright (C) 1996 U.N.S.A.
 *  
 *  xxl is a free software.  Permission to use, copy, and/or distribute
 *  this software and its documentation for any purpose and without fee is
 *  hereby granted, provided that both the above copyright notice and this
 *  permission notice appear in all copies and derived works. Fees for
 *  distribution or use of this software or derived works are forbidden.
 *  This software is provided ``as is'' without any warranty.
 *  
 *        Authors   : Fau Alexandre 
 *                    Enee Gilles 
 *                    Gaieb Mohamed Semi 
 *                    Palermo Gerald 
 * 
 *  Creation date   : 11-Jan-1996 17:48 
 *  Last file update: 21-Jun-1996 09:42 
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "booleen.h"
#include "stk.h"

#include "traducteur.h"

/* Prototypes */

static void Fill_Right_Parentheses(char **t);

static char *Double_Backslash(char *Ch);
static char *Suppresion_Blancs(char *Ch);
static char *ltoa(double l_to_a);
static char *d_toa(double d_to_a);
static char *Lit_Nom_Fichier(char **nf);

static char *Reconnait_Cellule(char **cell);
static char *Reconnait_Fonction(char **fct);
static char *Reconnait_Nombre(char **nbr);
static char *Reconnait_Symbole(char **symb);
static char *Reconnait_Chaine(char **chaine);

static char *Analyse_Prochain_Type(char **c);

PRIMITIVE STk_traducteur(SCM Chaine_Scheme);

static BOOLEEN Cellule_Figee(char car);
static BOOLEEN Cellule_Externe(char car);
static BOOLEEN Lettre(char car);
static BOOLEEN Chiffre(char car);
static BOOLEEN Symbole_Special(char car);
static BOOLEEN Chaine(char car);

/* comme ca le calcul n'est effectuee qu'une fois ! na na na nanere ! */
static int MIN2MAJ = 'a' - 'A';

/* Fonctions / Procdures */

static void Fill_Right_Parentheses(char **t) {

/* Rajoute les parentheses droites manquantes */

  long LPar = 0, RPar = 0;

  long i;

  char *temp = *t;

  while(*temp!='\0') {
    if(*temp=='(')
      LPar++;
    else if(*temp==')')
      RPar++;

    temp++;
  }

  if(RPar<LPar) { /* il manque des parentheses */

    for(i=RPar ; i<LPar ; i++) {
      strcat(*t,")");
    }

  } else if(RPar>LPar)
    strcpy(*t, "(list 'erreur)");
}

static char *Double_Backslash(char *Ch) {

  /* When returning a string to be evaluated by STK,
     the back slash are to be doubled to avoid evaluating
     problem : \" => \#" and there will be no string end. */

  char *Temp;
  char *Chaine_ret, *pt;

  if(Ch != NULL) {
    pt = Temp = calloc(strlen(Ch) * 2, sizeof(char));
    
    while((*Ch)!='\0') {
      *Temp = *Ch;
      if((*Ch)=='\\') {
	Temp ++;
	*Temp = '\\';

      } else if ((*Ch)=='"') { /* is also a problem */
	*Temp = '\\';
	Temp ++;
	*Temp = *Ch;
      }
      Temp ++;
      Ch ++;
    }
    *Temp = '\0';
    Chaine_ret = calloc(strlen(pt) + 1, sizeof(char));
    strcpy(Chaine_ret, pt);
  }
  else {
    Chaine_ret = calloc(1, sizeof(char));
    *Chaine_ret = '\0';
  }

  return Chaine_ret;
}

static char *Suppresion_Blancs(char *Ch) {

/* This function supresses blank caracter, trying not to
   change strings between "" and strings between ''. All
   non-protected strings are upcased.                    */

  char *Temp = calloc(strlen(Ch), sizeof(char));
  char *Chaine_ret = Temp;

  if (Ch!=NULL) {
    while(*Ch!='\0') {
      
      /* Si celui qui lit ce commentaire connait un truc plus simple
	 qu'il ne se gene pas et qu'il change ! */
      
      /* Traitement special pour les chaines */

      if (Chaine(*Ch)) {
	do {
	  *Temp = *Ch;
	  Temp++;
	  Ch++;
	} while(((*Ch)!='\0')&&(!Chaine(*Ch)));
      }

      if (Cellule_Externe(*Ch)) { /* Et pour les references externes */
	do {
	  *Temp = *Ch;
	  Temp++;
	  Ch++;
	} while(((*Ch)!='\0')&&(!Cellule_Externe(*Ch)));
      }

      /* Si c'est pas un blanc, on recopie */
      if ((*Ch != 8) && (*Ch != 32)){

	/* Et en plus on fait de la mise en majuscule ! */
	if((*Ch >= 'a') && (*Ch <= 'z')) {
	  *Temp = *Ch - MIN2MAJ;
	  Temp++;
	  Ch++;
	}
	else {
	  *Temp = *Ch;
	  Temp++;
	  Ch++;
	}
      }
      /* sinon, on saute les blancs */
      else {
	Ch ++;
      }
    }
  }
  *Temp='\0';
  return Chaine_ret;
}

/* Quelques prdicats de reconnaissance de la chaine tape */

static BOOLEEN Cellule_Figee(char car) {

/* Look for a $ caracter which indicates that the next cell value
   (column or lign) is absolute                                   */

  if (car == FIGEE)
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Cellule_Externe(char car) {

/* Tests if the current caracter is a begining of an external
   reference for a cell to a file name */

  if (car == DEBUT_NOM_FICHIER)
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Lettre(char car) {

/* Is car an alphabetic caracter ? */

  if ((car >= 'A') && (car <= 'Z'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Chiffre(char car) {

/* Is car a numeric caracter ? */

  if ((car >='0') && (car <= '9'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Symbole_Special(char car) {

/* Is car a special caracter ? */

  if ((car == SEPARATEUR) ||
      (car == DEF_ZONE) ||
      (car == '(') ||
      (car == ')') ||
      (car == '+') ||
      (car == '-') ||
      (car == '*') ||
      (car == '/') ||
      (car == '&') ||
      (car == '%'))
    return VRAI;
  else
    return FAUX;
}

static BOOLEEN Chaine(char car) {

/* Are we encountering a quotes string */

  if (car == '"')
    return VRAI;
  else
    return FAUX;
}

/* Les fonctions de "traduction" des diffrents types connus */

static char *ltoa(double l_to_a) {

/* Long to Ascii : using a double to transform great number ! 
   Returns the long integer in a string */

  char *ret_a = calloc(100, sizeof(char));

  sprintf(ret_a, "%.0f ", l_to_a);

  return ret_a;
}

static char *d_toa(double d_to_a) {

/* Double to Ascii : Converts a double into string */

  char *ret_a = calloc(100, sizeof(char));

  sprintf(ret_a, "%f ", d_to_a);

  return ret_a;
}

static char *Lit_Nom_Fichier(char **nf) {

/* This function return the file name between ' */

  char *pt = *nf;

  char *ret_nf = calloc(255, sizeof(char));
  char *ret = ret_nf;

  pt++;

  *ret_nf++ = '"';

  while((*pt) && (! Cellule_Externe(*pt)))
         *ret_nf++ = *pt++;

  if(! (*pt))
    return "'erreur";

  *ret_nf++ = '"';
  *ret_nf++ = ' ';
  *ret_nf   = '\0';

  *nf = ++pt;

  return ret;
}

static char *Reconnait_Cellule(char **cell) {

/* The purpose is to recognize a cell and to return it
   Scheme evaluable */

  long Colonne = 0;
  long Ligne = 0;

  char *Ret_Cell = calloc(200, sizeof(char));

  char *pt_cell = *cell;

  BOOLEEN Ferme_Feuille = FAUX;

  /* On reconnait d'abord le nom de la feuille externe s'il y a lieu */

  if(Cellule_Externe(*pt_cell)) {
    strcat(Ret_Cell, "(list 'Feuille ");
    strcat(Ret_Cell, Lit_Nom_Fichier(cell));
    pt_cell = *cell;
    Ferme_Feuille = VRAI;
  }

  /* On regarde si la colonne de la cellule est fige */

  strcat(Ret_Cell, "(list ");

  if(Cellule_Figee(*pt_cell)) {
    strcat(Ret_Cell, "'$ ");
    pt_cell++;
    *cell = pt_cell;
  }
  else
    strcat(Ret_Cell, "'r ");

  /* Reconnaissance de la cellule */

  /* Colonne */

  while((*pt_cell) && Lettre(*pt_cell)) {
    Colonne *= 26;
    Colonne += ((*pt_cell++) - 'A' + 1);
  }

  *cell = pt_cell;

  if(Colonne>0)
         strcat(Ret_Cell, ltoa((double) Colonne));
  else
         strcat(Ret_Cell, "'erreur");

  /* On regarde si la ligne de la cellule est fige */

  if(Cellule_Figee(*pt_cell)) {
    strcat(Ret_Cell, "'$ ");
    pt_cell++;
    *cell = pt_cell;
  }
  else
    strcat(Ret_Cell, "'r ");

  /* Ligne */
  
  while((*pt_cell) && Chiffre(*pt_cell)) {
    Ligne *= 10;
    Ligne += (*pt_cell++) - '0';
  }

  *cell = pt_cell;

  if(Ligne>0)
         strcat(Ret_Cell, ltoa((double) Ligne));
  else
         strcat(Ret_Cell, "'erreur");

  strcat(Ret_Cell, ")");

  if (Ferme_Feuille)
         strcat(Ret_Cell, ")");

  return Ret_Cell;
}

static char *Reconnait_Fonction(char **fct) {

  char *pt_nfct = *fct;

  char *Ret_Fct = calloc(1000, sizeof(char));

  char *Ret_Tmp = Ret_Fct;

  strcat(Ret_Fct, "(list '");
  
  Ret_Fct += strlen(Ret_Fct); 

  /* Lecture du symbole de la fonction */

  while(Lettre(*pt_nfct))
         *Ret_Fct++ = *pt_nfct++;

  *Ret_Fct++ = ' ';
  *Ret_Fct = '\0';

  Ret_Fct = Ret_Tmp; /* On remet le pointeur a retourner au bon endroit */
  
  *fct = pt_nfct;

  if (*pt_nfct != '(') { /* Pas de paramertres */
    strcat(Ret_Fct, ")");
    return Ret_Fct;
  }
  else {
    (*fct)++;

    while((**fct)!='\0' && (**fct) != ')')
      strcat(Ret_Fct, Analyse_Prochain_Type(fct));
    
    strcat(Ret_Fct, " )");
    (*fct)++;
  }
  return Ret_Fct;

}

static char *Reconnait_Nombre(char **nbr) {

/* Recognize a number : integer or double */

  char *pt_nb = *nbr;
  
  double Nombre = 0.;
  double Puissance = 1.;

  while(Chiffre(*pt_nb)) {
    Nombre *= 10;
    Nombre += (*pt_nb++) - '0';
  }
  
  if((*pt_nb)==DECIMALE) {
    pt_nb++;
    while(Chiffre(*pt_nb)) {
      Puissance *= 10;
      Nombre *= 10;
      Nombre += (*pt_nb++) - '0';
    }
    Nombre /= Puissance;
    *nbr = pt_nb;
    
    return d_toa(Nombre);

  } else {
    *nbr = pt_nb;
    
    return ltoa(Nombre);
  }
}

static char *Reconnait_Symbole(char **symb) {

/* Converts special symbols into Scheme easy symbol */
  
  char car = **symb;

  char *Ret_Symb = calloc(5, sizeof(char));

  Ret_Symb[0] = ' ';
  Ret_Symb[1] = '\'';
  Ret_Symb[2] = car;
  Ret_Symb[3] = ' ';

  if (car == ';') {   /* ; is a special separator in scheme so we use \ */
    Ret_Symb[1]='\'';
    Ret_Symb[2]='\\';
  }

  (*symb)++;

  if (car == '(')    /* Begining of a new list */
    return "(list ";

  if (car == ')')
    Ret_Symb[1]=' ';

  return Ret_Symb;
}

static char *Reconnait_Chaine(char **chaine) {

/* Returns a string between "" in an evaluable scheme string */

  char *Ret_Chaine = calloc(255, sizeof(char));
  char *pt_chaine = *chaine;
  char *pt_ret = Ret_Chaine;

  *Ret_Chaine++ = *pt_chaine++;

  while(!(*pt_chaine == '\0') && !Chaine(*pt_chaine))
    *Ret_Chaine++ = *pt_chaine++;

  if(*pt_chaine == '\0')
    return "'erreur";

  *Ret_Chaine = *pt_chaine++;

  *chaine = pt_chaine;

  return pt_ret;
}

/* Fonction d'analyse du prochaine type */

static char *Analyse_Prochain_Type(char **c) {

/* The main analysis function */

  char *pt_c = *c;
  char *pt_t;

  /* Cellule ? */

  if((Cellule_Figee(*pt_c)) ||  /* A priori un $ */
     (Cellule_Externe(*pt_c)))  /* une ' pour le nom de la feuille */
    return Reconnait_Cellule(c);

  /* Cellule ou fonction ? */

  if(Lettre(*pt_c)) { /* Soit fonction, soit cellule */
    pt_t = pt_c;
    while(Lettre(*pt_t)) pt_t++;
    if(Chiffre(*pt_t)||Cellule_Figee(*pt_t)) 
      return Reconnait_Cellule(c);
    else 
      return Reconnait_Fonction(c);
   
  }

  /* Nombre ? */

  if((Chiffre(*pt_c)) || (*pt_c) == DECIMALE) 
    return Reconnait_Nombre(c);

  /* Symbole ? */

  if(Symbole_Special(*pt_c)) 
    return Reconnait_Symbole(c);
  /* Chaine de caractres */

  if(Chaine(*pt_c)) 
    return Reconnait_Chaine(c);

  (*c)++;

  /* Type non reconnu */
  return "'erreur";

}

PRIMITIVE STk_traducteur(SCM Chaine_Scheme) {

  /* STk called function which only transform scheme data into
     C usable feature. */

  /* Chaine de dpart */
  char *Chaine_C = Chaine_Scheme->storage_as.string.data;

  char *PT;
  SCM res;

  /* Chaine de travail */
  char *Chaine_T;

  /* Chaine servant  l'valuation du rsultat */
  char *Chaine_R = calloc(1000, sizeof(char));

  /* Analyse de la chaine */

  /* Suppression des blancs et recopie de Chaine_C : Raciste ! */

  Chaine_T = Suppresion_Blancs(Chaine_C);

  /* On remplit le dbut de la chaine qui sert au rsultat */

  strcpy(Chaine_R, "(list ");

  /* On regarde d'abord s'il y a quelque chose a faire ! */

  if (*Chaine_T != '=') {

    /* On regarde si on a un nombre saisi */

    Chaine_C = Double_Backslash(Chaine_C);

    if (Chiffre(*Chaine_T)) {
      PT = Reconnait_Nombre(&Chaine_T);
      if(*Chaine_T == '\0') { /* plus joli */
	strcat(Chaine_R, "'no_eval ");
	strcat(Chaine_R, PT);
	strcat(Chaine_R, ")");	
	return STk_eval_C_string(Chaine_R, UNBOUND);
      }
      else {
	
	/* On met 'no_eval dans la liste retourne
	   pour que le moteur sache qu'il n'y a rien a faire */
	
	strcat(Chaine_R, "'no_eval ");
	strcat(Chaine_R, "\""); /* \" pb lors de la transformation en chaine scheme */
	strcat(Chaine_R, Chaine_C);
	strcat(Chaine_R, "\")");
	
	return STk_eval_C_string(Chaine_R, UNBOUND); /* @@@@@ */
      }
    }
    else {

      /* On met 'no_eval dans la liste retourne
       pour que le moteur sache qu'il n'y a rien a faire */

      strcat(Chaine_R, "'no_eval ");
      strcat(Chaine_R, "\""); /* \" pb lors de la transformation en chaine scheme */
      strcat(Chaine_R, Chaine_C);
      strcat(Chaine_R, "\")");

      return STk_eval_C_string(Chaine_R, UNBOUND); /* @@@@@ */
    }
  }

  /* C'est bien un gal, on doit analyser la suite */

  Chaine_T++;
  do {
    PT = Analyse_Prochain_Type(&Chaine_T);
    strcat(Chaine_R, PT);
  } while (((*Chaine_T) != '\0') && (strstr(PT, "'erreur") == NULL));

  strcat(Chaine_R, ")");
  Chaine_R[strlen(Chaine_R)+1]='\0';

  Fill_Right_Parentheses(&Chaine_R);
 
  if  (strstr(PT, "'erreur") != NULL)
    return STk_eval_C_string("(list 'erreur ) " , UNBOUND); /* @@@@@ */
  else
    return STk_eval_C_string(Chaine_R , UNBOUND); /* @@@@@ */
}

PRIMITIVE STk_init_traducteur(void) {

  /* The init function called when you load "traducteur.so" under STk */

  STk_add_new_primitive("traducteur", tc_subr_1, STk_traducteur);

  return UNDEFINED;
}
