/*
;;;; tradinv.c
;;;;
;;;; xxl project -- Universite de Nice Sophia Antipolis
;;;;                U.F.R. Sciences - Maitrise d'Informatique
;;;;                project supervisor  V. Granet (vg@unice.fr)
;;;;
;;;; Copyright (C) 1996 U.N.S.A.
;;;; 
;;;; xxl is a free software.  Permission to use, copy, and/or distribute
;;;; this software and its documentation for any purpose and without fee is
;;;; hereby granted, provided that both the above copyright notice and this
;;;; permission notice appear in all copies and derived works. Fees for
;;;; distribution or use of this software or derived works are forbidden.
;;;; This software is provided ``as is'' without any warranty.
;;;; 
;;;;       Authors   : Fau Alexandre 
;;;;                   Enee Gilles 
;;;;                   Gaieb Mohamed Semi 
;;;;                   Palermo Gerald 
;;;;
;;;; Creation date   : 11-Jan-1996 17:48 
;;;; Last file update:  1-Jul-1996 15:07 
*/
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "stk.h"

#include "booleen.h"

/* Prototypes */

static BOOLEEN
Ref_ExterneP(SCM Liste);

static char *
Reconnait_Ref_Externe(SCM Liste);

static BOOLEEN
SeparatorP(char *Arg_Lu);

static char *
Reconnait_Fonction(SCM Liste);

static BOOLEEN
NomFonctionP(char *Symbole_vu);

static BOOLEEN
FonctionP(SCM Liste);

static char *
Reconnait_Cellule(SCM Liste);

static BOOLEEN
RelAbsP(SCM Objet);

static BOOLEEN
CelluleP(SCM Liste);

static char *
NumToAscii(long Colonne);

static BOOLEEN
AbsoluP(char *Chaine);

static char *
Analyse_Liste(SCM Liste);

PRIMITIVE 
STk_tradinv(SCM Liste_Gege, int argcount);

/* Fonctions / Procdures */

static BOOLEEN First_level, Second_level, Ref_Externe;

static BOOLEEN
Ref_ExterneP(SCM Liste) {

  return (strcmp(PNAME(Liste), "feuille")==0);
}

static char*
Reconnait_Ref_Externe(SCM Liste) {

  /* An external reference is a file name and a cell reference */

  char *Ret_Chaine = calloc(500, sizeof(char));

  Liste = CDR(Liste);

  sprintf(Ret_Chaine, "'%s'", CHARS(CAR(Liste)));

  Liste = CDR(Liste);

  strcat(Ret_Chaine, Reconnait_Cellule(CAR(Liste)));

  return Ret_Chaine;
}

static BOOLEEN
SeparatorP(char *Arg_Lu) {

  return (*Arg_Lu=='\\');
}

static char *
Reconnait_Fonction(SCM Liste) {

  char *Ret_Chaine = calloc(500, sizeof(char));

  strcat(Ret_Chaine, PNAME(CAR(Liste)));

  strcat(Ret_Chaine, "(");
  while(NNULLP(CDR(Liste))) { /* Analyse des arguments */
    Liste = CDR(Liste);

    if (SYMBOLP(CAR(Liste)) && SeparatorP(PNAME(CAR(Liste))))
      strcat(Ret_Chaine, ";");
    else
      strcat(Ret_Chaine, Analyse_Liste(CAR(Liste)));
  }
  strcat(Ret_Chaine, ")");
  
  return Ret_Chaine;
}

static BOOLEEN
Lettre(char car) {
  return (((car >= 'A') && (car <= 'Z')) ||
	  ((car >= 'a') && (car <= 'z')));
}

static BOOLEEN
NomFonctionP(char *Symbole_vu) {

  /* Dit si Symbole_vu est un nom de fonction : Lettre+ */

  if(! Lettre(*Symbole_vu))
    return FAUX;

  while((*Symbole_vu!='\0') && (Lettre(*Symbole_vu)))
    Symbole_vu ++;

  return (*Symbole_vu == '\0');
}

static BOOLEEN
FonctionP(SCM Objet) {

  /* Reconnait si nous avons a faire a une fonction */

  return (SYMBOLP(Objet) &&
	  NomFonctionP(PNAME(Objet)));
}

static char *
Reconnait_Cellule(SCM Liste) {

  /* Reconnait une cellule '(r 1 $ 1) ==> "A1 " */

  char *Ret_Chaine = calloc(500, sizeof(char));

  char *Recup;

  Recup = Analyse_Liste(CAR(Liste));
  if (AbsoluP(Recup))
    strcat(Ret_Chaine, "$");
  strcat(Ret_Chaine, NumToAscii(INTEGER(CAR(CDR(Liste)))));
  Recup = Analyse_Liste(CAR(CDR(CDR(Liste))));
  if (AbsoluP(Recup))
    strcat(Ret_Chaine, "$");
  Recup = Analyse_Liste(CAR(CDR(CDR(CDR(Liste)))));
  strcat(Ret_Chaine, Recup);

  return Ret_Chaine;
}

static BOOLEEN
RelAbsP(SCM Objet) {
  return (SYMBOLP(Objet) &&
	  (strcmp(PNAME(Objet), "r")==0 || strcmp(PNAME(Objet), "$")==0));
}

static BOOLEEN
CelluleP(SCM Liste){

  return (RelAbsP(CAR(Liste)) && 
	  RelAbsP(CAR(CDR(CDR(Liste)))) &&
	  INTEGERP(CAR(CDR(Liste))) &&
	  INTEGERP(CAR(CDR(CDR(CDR(Liste))))));
}

static char *
NumToAscii(long Colonne){

  /* Convertit un numero de colonne en sa "lettre" 

      1 ==> "A" 
     27 ==> "AA" 

     Pas mal pour les plaques d'immatriculation ! */

  char *Ret_Chaine = calloc(30, sizeof(char));

  char *pt = Ret_Chaine + (28 * sizeof(char));

  char *tpt;

  char reste;

  if(Colonne > 26) {

    while(Colonne>25) {
    
      Colonne --;
      
      reste = Colonne - ((Colonne / 26) * 26);
    
      Colonne /= 26;

      *pt = reste + 'A';

      pt --;
    }

    if(Colonne>0)
      *pt = (char) Colonne + 'A' - 1;
    else
      pt ++;

  } else
    *pt = (char) Colonne + 'A' - 1;

  return pt;
}

static BOOLEEN
AbsoluP(char *Chaine) {
  return (*Chaine=='$');
}

static char *
Analyse_Liste(SCM Liste) {

  char *Ret_Chaine = calloc(500, sizeof(char));

  char *Recup;

  BOOLEEN Parentheses = FAUX;

  /* On va faire une sorte de cond pour analyser la liste
     comme nous l'avons fait en scheme cette annee pour
     le Mini Blaise entre autre, mais en C !!!            */

  if(NULLP(Liste))
    return "";

  else if (SYMBOLP(Liste) && SeparatorP(PNAME(Liste))) /* Cas du ; */
    sprintf(Ret_Chaine, ";");

  else if (SYMBOLP(Liste))
    sprintf(Ret_Chaine, "%s", PNAME(Liste)); /* Symbole quelconque */

  else if (INTEGERP(Liste)) {
    sprintf(Ret_Chaine, "%ld", INTEGER(Liste)); /* Entier */

  } else if (BIGNUMP(Liste)) {
    sprintf(Ret_Chaine, "%ld", BIGNUM(Liste)); /* Pas terrible ce truc */

  } else if (FLONUMP(Liste)) {
    sprintf(Ret_Chaine, "%f", FLONM(Liste)); /* Les doubles */

  } else if (STRINGP(Liste) && Ref_Externe) {
    sprintf(Ret_Chaine, "'%s'", CHARS(Liste));
    Ref_Externe = FAUX;

  } else if (STRINGP(Liste)) {
    sprintf(Ret_Chaine, "\"%s\"", CHARS(Liste)); /* Les chaines entre "" */

  } else if (KEYWORDP(Liste)) {
    sprintf(Ret_Chaine, ":");

  } else if(CelluleP(Liste))
    Ret_Chaine = Reconnait_Cellule(Liste); /* Les cellules */

  else if (FonctionP(CAR(Liste)) && Ref_ExterneP(CAR(Liste))) {
    Ret_Chaine = Reconnait_Ref_Externe(Liste);

  } else if(FonctionP(CAR(Liste))) {
    Ret_Chaine = Reconnait_Fonction(Liste); /* Les fonctions */

  } else {
    if(First_level)
      First_level = FAUX;
    else if(Second_level)
      Second_level = FAUX;
    else {
      Parentheses = VRAI;
      strcpy(Ret_Chaine, "(");
    }
    do {  /* Reste les listes ! */
      Recup = Analyse_Liste(CAR(Liste));
      strcat(Ret_Chaine, Recup);
      Liste = CDR(Liste);
    } while(NNULLP(Liste));
    if(Parentheses)
      strcat(Ret_Chaine, ")");
  }
  return Ret_Chaine;
}

PRIMITIVE 
STk_tradinv(SCM Liste_Gege, int argcount) {

  char *ret_str = calloc(500, sizeof(char));

  /* Le but est de prendre une liste du type de celle
     que l'on a a la sortie du traducteur et de la
     remettre en chaine pour le COPIER/COLLER.        */

  First_level = VRAI;
  Second_level = VRAI;
  Ref_Externe = FAUX;

  if(argcount > 0) {
    strcat(ret_str, "=");
    strcat(ret_str, Analyse_Liste(Liste_Gege));
    return STk_makestring(ret_str);

  } else
    return STk_makestring("");

}

PRIMITIVE 
STk_init_tradinv(void) {

  STk_add_new_primitive("tradinv", tc_lsubr, STk_tradinv);

  return UNDEFINED;
}
